/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Copyright (C) 2021-2024 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::fv::phaseChange

Description
    Base class for phase change models

SourceFiles
    phaseChange.C

\*---------------------------------------------------------------------------*/

#ifndef phaseChange_H
#define phaseChange_H

#include "massTransfer.H"
#include "ThermoRefPair.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

class basicThermo;
class fluidThermo;
class multicomponentThermo;

namespace fv
{

/*---------------------------------------------------------------------------*\
                          Class phaseChange Declaration
\*---------------------------------------------------------------------------*/

class phaseChange
:
    public massTransfer
{
private:

    // Private Data

        //- The thermo references
        const ThermoRefPair<basicThermo> thermos_;

        //- The fluid thermo references
        const ThermoRefPair<fluidThermo> fluidThermos_;

        //- The specie thermo references
        const ThermoRefPair<multicomponentThermo> specieThermos_;

        //- Names of the energy fields
        const Pair<word> heNames_;


protected:

    // Protected Member Functions

        //- Access the pressure field
        const volScalarField& p() const;

        //- Add a boundary field to the given internal field
        static tmp<volScalarField> vifToVf
        (
            const tmp<DimensionedField<scalar, volMesh>>& tvif
        );

        //- Remove the boundary field from the given geometric field
        static tmp<DimensionedField<scalar, volMesh>> vfToVif
        (
            const tmp<volScalarField>& tvf
        );


public:

    //- Runtime type information
    TypeName("phaseChange");


    // Constructors

        //- Construct from explicit source name and mesh
        phaseChange
        (
            const word& name,
            const word& modelType,
            const fvMesh& mesh,
            const dictionary& dict,
            const Pair<bool>& fluidThermosRequired,
            const Pair<bool>& specieThermosRequired
        );


    // Member Functions

        // Access

            //- Return the thermo references
            inline const ThermoRefPair<basicThermo>& thermos() const;

            //- Return the fluid thermo references
            inline const ThermoRefPair<fluidThermo>& fluidThermos() const;

            //- Return the specie thermo references
            inline const ThermoRefPair<multicomponentThermo>&
                specieThermos() const;

            //- Return the names of the energy fields
            inline const Pair<word>& heNames() const;


        // Evaluation

            //- Return the temperature at which the phases are considered to be
            //  changing. By default this is considered to be the temperature
            //  of the "source" phase (i.e., the phase for which mDot is
            //  negative), but this can be overridden to account for heat
            //  transfer modelling or similar.
            virtual tmp<DimensionedField<scalar, volMesh>> Tchange() const;

            //- Return the fraction of the latent heat that is transferred into
            //  the second phase. By default this is weighted by the phase
            //  thermal conductivities, but this can be overridden to account
            //  for heat transfer modelling or similar.
            virtual tmp<DimensionedField<scalar, volMesh>> Lfraction() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace fv
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "phaseChangeI.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
